//Authors: Dario Cattaruzza, Alessandro Abate, Peter Schrammel, Daniel Kroening
//University of Oxford 2016
//This code is supplied under the BSD license agreement (see license.txt)

#ifndef SPARSESORTEDMATRIX_H
#define SPARSESORTEDMATRIX_H

#include "Scalar.h"
#include "Set.h"
#include "RowSort.h"
#include <vector>

namespace abstract{

template <class scalar>
class SparseSortedMatrix : public interval_def<scalar>::sparseMatrix, public SortedTableau<scalar> {
public:
    typedef interval_def<scalar> type;
    typedef typename type::scalar_type refScalar;
    typedef functions<refScalar> func;
    typedef typename type::sparseVector sparseVector;
    typedef typename type::sparseMatrix sparseMatrix;
    typedef typename type::MatrixS MatrixS;

    using sparseMatrix::coeff;
    using sparseMatrix::coeffRef;
    using sparseMatrix::rows;
    using sparseMatrix::cols;
    using sparseMatrix::resize;
    using sparseMatrix::conservativeResize;

    using SortedTableau<scalar>::m_dirty;
    using SortedTableau<scalar>::m_order;
    using SortedTableau<scalar>::order;
    using SortedTableau<scalar>::zeroOrder;
    using SortedTableau<scalar>::ComputeRowOrderVector;

    SparseSortedMatrix(int rows,int cols);

    SparseSortedMatrix(sparseMatrix &source,const OrderType &type=LexNone,bool transpose=false,bool norm=false);

    char compareZeroOrderRows(int row1,int row2);
    scalar diffZeroOrderRows(int row1,int row2);
    scalar diffZeroOrderRows(int row1,int row2,MatrixS &ranges);

    void QuickAngleSort();

    int numRows()                         { return rows(); }

    int numCols()                         { return cols(); }

    // Sets the size of the tableau
    void setSize(int rows,int cols)       { conservativeResize(rows,cols); }

    // Gets a given coefficient in the sorted Tableau
    const scalar getCoeff(int row,int col)              { return coeff(row,col); }

    // Sets a given coefficient in the sorted Tableau
    void setCoeff(int row,int col,const scalar &src)    { coeffRef(row,col)=src; }

    // Sets the given row to src starting at col
    void setRow(int row,int col,const MatrixS &src);

    // Sets the given row to -src starting at col
    void negRow(int row,int col,const MatrixS &src);

    // Substracts the given row from dest
    void subRow(MatrixS &dest,int row);

    // Normalizes each row
    void normalize();

    /// Takes the ordering from  a different sort (eg a normalized one)
    bool copyOrder(SparseSortedMatrix<scalar> &src);

    void sortCols();

    /// Pivots the tableau in place on the given pivot
    void pivot(MatrixS &auxiliary,pivot_t pivot);

    /// Retrieves the value of a tableau entry for a given row and basis column (see revised simplex algorithm)
    scalar entry(MatrixS &basis,const int row, const int col);

    ///Retrieves the magnitude of the vector in the direction normal to the given halfspace
    ///@param vector vector to verify
    ///@param row halfspace to check against
    scalar entry(const MatrixS &vector,const int row);
protected:
    /// Checks if row1<row2 ordered by adjacent column angle
    /// @param row1 row to check for
    /// @param row2 row to check against
    /// @return true if row1 is smaller than row2 by angle
    bool SmallerAngleRow(const int row1,const int row2);

    /// Checks if row1>row2 ordered by adjacent column angle
    /// @param row1 row to check for
    /// @param row2 row to check against
    /// @return true if row1 is larger than row2 by angle
    bool LargerAngleRow(const int row1,const int row2);

    /// Checks if row1<row2 ordered by column
    /// @param row1 row to check for
    /// @param row2 row to check against
    /// @return true if row1 is smaller than row2
    bool SmallerRow(const int row1,const int row2);

    /// Checks if row1>row2 ordered by column
    /// @param row1 row to check for
    /// @param row2 row to check against
    /// @return true if row1 is larger than row2
    bool LargerRow(const int row1,const int row2);

    void cosineSort();

    bool SmallerCos(std::vector<scalar> &cosines,const int row1,const int row2);

    bool LargerCos(std::vector<scalar> &cosines,const int row1,const int row2);

    long subCosPartition(std::vector<int> &order, std::vector<scalar> &cosines, const long p, const long r);

    void QuickCosSort(std::vector<int> &order, std::vector<scalar> &cosines, const long p, const long r);

    void cosSort(std::vector<int> &order, std::vector<scalar> &cosines);

    refScalar maxTightWidth(const MatrixS &vector, const int row);

    refScalar maxTightWidth(const MatrixS &basis, const int row,const int col);

    /// Same as above but with tighter bounds
    scalar tightEntry(MatrixS &basis,const int row, const int col,const int iteration=0);

    ///Retrieves the magnitude of the vector in the direction normal to the given halfspace
    ///@param vector vector to verify
    ///@param row halfspace to check against
    inline scalar tightEntry(const MatrixS &vector, const long row);

    ///Relaxes the current basis by the given interval to cope with imprecisions
    void relaxBasis(MatrixS &basis,scalar width,int col);
public:
    std::vector<refScalar>      m_cosines;
    MatrixS                     m_norms;
};

}
#endif
