//Authors: Dario Cattaruzza, Alessandro Abate, Peter Schrammel, Daniel Kroening
//University of Oxford 2016
//This code is supplied under the BSD license agreement (see license.txt)
#include "CanonicalSystem.h"
#include <boost/timer.hpp>
#include <set>
#include <Eigen/QR>
#include <Eigen/Eigenvalues>

namespace abstract{

template <class scalar>
canonicalType_t CanonicalSystem<scalar>::ms_canonical=eNNF;

/// Constructs an empty buffer
template<class scalar>
CanonicalSystem<scalar>::CanonicalSystem(int dimension,int idimension,int odimension,bool trace) :
  DigitalSystem<scalar>(dimension,idimension,odimension,trace),
  m_T(0,0),
  m_observerDynamics(0,0),
  m_observerSensitivity(0,0),
  m_observerOutputSensitivity(0,0),
  m_observerDynamicsError(0),
  m_observerSensitivityError(0),
  m_observerOutputSensitivityError(0)
{}

/// Copies an existing object
template<class scalar>
void CanonicalSystem<scalar>::copy(const CanonicalSystem &source)
{
  DigitalSystem<scalar>::copy(source);
  m_observer=source.m_observer;
  m_observerDynamics=source.m_observerDynamics;
  m_observerSensitivity=source.m_observerOutputSensitivity;
  m_observerOutputSensitivity=source.m_observerOutputSensitivity;
  m_observerDynamicsError=source.m_observerDynamicsError;
  m_observerSensitivityError=source.m_observerSensitivityError;
  m_observerOutputSensitivityError=source.m_observerOutputSensitivityError;
}

/// Transforms the matrix to Row Echelon Form
template <class scalar>
int CanonicalSystem<scalar>::toREF(MatrixS &matrix)
{
  int rank=m_dimension;
  int col=0;
  for (int row=0;col<matrix.rows();row++,col++) {
    while (func::isZero(func::norm2(matrix.coeff(row,col)))) {
      for (int row2=row+1;row2<matrix.rows();row2++) {
        if (!func::isZero(func::norm2(matrix.coeff(row2,col)))) {
          matrix.row(row)+=matrix.row(row2);
          break;
        }
      }
      if (func::isZero(func::norm2(matrix.coeff(row,col)))) {
        col++;
        if (col==matrix.cols()) break;
      }
    }
    if (col==matrix.cols()) break;
    scalar multiplier=matrix.coeff(row,col);
    if (!func::isZero(func::norm2(multiplier))) {
      matrix.row(row)/=multiplier;
      rank--;
    }
    for (int row2=row+1;row2<matrix.rows();row2++) {
      scalar multiplier=matrix.coeff(row2,col);
      matrix.row(row2)-=multiplier*matrix.row(row);
    }
  }
  if (ms_trace_dynamics[eTraceREF]) ms_logger.logData(matrix,"REF:");
  return rank;
}

/// Retrieves the characteristic polynomial coefficients of the dynamics
template <class scalar>
typename CanonicalSystem<scalar>::MatrixS CanonicalSystem<scalar>::getDynamicPolynomialCoefficients(const MatrixC &eigenValues)
{
  int dimension=eigenValues.rows();
  MatrixC result=MatrixC::Zero(2,dimension+1);
  result.coeffRef(0,0)=func::ms_1;
  result.coeffRef(0,1)=-eigenValues.coeff(0,0);
  for (int i=1;i<dimension;i++) {
    result.row(1)=-result.row(0)*eigenValues.coeff(i,i);
    result.block(0,1,1,i+1)+=result.block(1,0,1,i+1);
  }
  return result.real();
}

/// Retrieves the characteristic polynomial coefficients of the dynamics
template <class scalar>
typename CanonicalSystem<scalar>::MatrixS CanonicalSystem<scalar>::getDynamicPolynomialCoefficients()
{
  return getDynamicPolynomialCoefficients(m_eigenValues);
}

/// Retrieves the reachability matrix [B AB A^2B ...A^{n-1}B]
template <class scalar>
typename CanonicalSystem<scalar>::MatrixS CanonicalSystem<scalar>::getReachabilityMatrix()
{
  MatrixS result(m_sensitivity.rows(),m_dimension*m_sensitivity.cols());
  result.block(0,0,m_sensitivity.rows(),m_sensitivity.cols())=m_sensitivity;
  MatrixS multiplier=m_dynamics;
  for (int i=1;i<m_dimension;i++)
  {
    result.block(0,i*m_sensitivity.cols(),m_sensitivity.rows(),m_sensitivity.cols())=multiplier*m_sensitivity;
    multiplier*=m_dynamics;
  }
  m_controllabilityIndices.resize(m_fdimension,0);
  if (m_fdimension>1) {
    MatrixS matrix=result;
    int rank=toREF(matrix);
    int col=0;
    for (int row=0;row<rank;row++) {
      while (func::isZero(matrix.coeff(row,col))) col++;
      result.col(row)=result.col(col);
      m_controllabilityIndices[col%m_fdimension]++;
    }
    result.conservativeResize(result.rows(),result.rows());
  }
  else m_controllabilityIndices[0]=m_dimension;
  if (ms_trace_dynamics[eTraceController]) ms_logger.logData(result,"Reachability Matrix");
  return result;
}

/// Retrieves the reachability matrix [1 a1 a2...a_{n-1};0 1 a1 ... a_{n-2}...]
template <class scalar>
typename CanonicalSystem<scalar>::MatrixS CanonicalSystem<scalar>::getCanonicalReachabilityMatrix()
{
  //TODO: not sure how this works on MIMO
  MatrixS result=MatrixS::Identity(m_sensitivity.rows(),m_dimension*m_sensitivity.cols());
  MatrixS coefficients=getDynamicPolynomialCoefficients();
  if (ms_trace_dynamics[eTraceController]) {
    MatrixS matrix=coefficients.row(0);
    ms_logger.logData(matrix,"Coefficients");
  }
  for (int i=1;i<m_dimension;i++) result.block(i-1,i,1,m_dimension-i)=coefficients.block(0,1,1,m_dimension-i);
  if (ms_trace_dynamics[eTraceController]) ms_logger.logData(result,"Canonical Reachability Matrix");
  return result;
}

/// Retrieves the transform matrix T : z=T^{-1}x turns A,B,C into controllable canonical form
template <class scalar>
typename CanonicalSystem<scalar>::MatrixS CanonicalSystem<scalar>::getReachableCanonicalTransformMatrix()
{
  MatrixS result=getReachabilityMatrix()*getCanonicalReachabilityMatrix();
  return makeInverse(result);
}

/// Retrieves the observability matrix [C CA CA^2 ...CA^{n-1}]^T
template <class scalar>
typename CanonicalSystem<scalar>::MatrixS CanonicalSystem<scalar>::getObservabilityMatrix()
{
  MatrixS outputSensitivity=m_outputSensitivity;
  if (outputSensitivity.cols()==0) {
    outputSensitivity=MatrixS::Zero(1,m_dimension);
    outputSensitivity.coeffRef(0,0)=func::ms_1;
  }
  MatrixS result(m_dimension*outputSensitivity.rows(),outputSensitivity.cols());
  result.block(0,0,outputSensitivity.rows(),outputSensitivity.cols())=outputSensitivity;
  MatrixS multiplier=m_dynamics;
  for (int i=1;i<m_dimension;i++)
  {
    result.block(i*outputSensitivity.rows(),0,outputSensitivity.rows(),outputSensitivity.cols())=outputSensitivity*multiplier;
    multiplier*=m_dynamics;
  }
  return result;
}

/// Retrieves the transform matrix T : z=T^{-1}x turns A,B,C into observable canonical form
template <class scalar>
typename CanonicalSystem<scalar>::MatrixS CanonicalSystem<scalar>::getObservableCanonicalTransformMatrix()
{
  MatrixS CanonicalObservabilityMatrix=getCanonicalReachabilityMatrix().transpose();
  MatrixS ObservabilityMatrix=getObservabilityMatrix();
  return makeInverse(ObservabilityMatrix)*makeInverse(CanonicalObservabilityMatrix);
  CanonicalObservabilityMatrix*=ObservabilityMatrix;
  return makeInverse(CanonicalObservabilityMatrix);
}

/// Creates matrices T, invT , [W, invW], to make a reachable [observable] system
template <class scalar>
void CanonicalSystem<scalar>::makeReachabilityMatrices(bool observer)
{
/*  MatrixS canonical=getCanonicalReachabilityMatrix();
  m_invT=getReachabilityMatrix()*canonical;
  m_T=makeInverse(m_invT);
  if (observer) {
    m_W=canonical*getObservabilityMatrix();
    m_invW=makeInverse(m_W);
  }*/
  m_T=getReachableCanonicalTransformMatrix();
  m_invT=makeInverse(m_T);
  if (observer) {
    m_invW=getObservableCanonicalTransformMatrix();
    m_W=makeInverse(m_invW);
  }
}

/// Creates bounds for K,L, KT_c^-1 and T_oL
template <class scalar>
void CanonicalSystem<scalar>::makeControlBounds(bool observer,int observerBits)
{
  if (m_T.rows()==0) makeReachabilityMatrices(observer);
  m_TT=m_T.transpose();
  m_invTT=m_invT.transpose();
  MatrixS controllable=m_T*m_dynamics*m_invT;
  if (ms_trace_dynamics[eTraceController]) {
    ms_logger.logData(controllable,"Controllable");
  }
  m_polynomialCoefficients=-controllable.row(0);
  AbstractPolyhedra<scalar> K=getControllerInBounds(/*m_safeReachTube*/m_initialState.getPolyhedra());
  AbstractPolyhedra<scalar> KS=getControllerStabilityBounds(m_polynomialCoefficients,false);
  KS.mirror();
  AbstractPolyhedra<scalar> KSC(KS);
  KSC.transform(m_TT,m_invTT);
  K.intersect(KSC);
  AbstractPolyhedra<scalar> L(KS);
  L.transform(m_invW,m_W);
  if (observerBits>0) {
    AbstractPolyhedra<scalar> Lbits(m_dimension);
    Lbits.loadData("cube<"+ms_logger.MakeNumber(1<<observerBits));
    L.intersect(Lbits);
  }
  K.normalise(true);
  K.toInner(true);
  m_controllerBounds.load(K);
  m_controllerBounds.setEigenSpace(m_TT,m_invTT);
  m_controllerBounds.getPolyhedra(eEigenSpace).normalise(true);
  if (observer) {
    L.normalise();
    L.toOuter(true);
    m_observerBounds.load(L);
    m_observerBounds.setEigenSpace(m_invW,m_W);
    m_observerBounds.getPolyhedra(eEigenSpace).normalise(true);
  }
}

/// Retrieves the reference gain
template <class scalar>
typename CanonicalSystem<scalar>::MatrixS CanonicalSystem<scalar>::getReferenceGain()
{
  MatrixS closedLoop=m_dynamics-(m_sensitivity*m_feedback);
  return m_outputSensitivity*makeInverse(closedLoop)*m_sensitivity;
}

/// Corrects the support set by the input offset
template <class scalar>
void CanonicalSystem<scalar>::demergeAccelInSupports(MatrixS &supports,MatrixS &inSupports,int numTemplates)
{
  if (!m_hasOnes || (m_inputType>=eVariableInputs))  {
    for (int row=0;row<numTemplates;row++) {
      int pos=row*m_numVertices;
      supports.coeffRef(pos,0)-=inSupports.coeff(0,row);
      for (int point=1;point<m_numVertices;point++) {
        supports.coeffRef(pos+point,0)-=inSupports.coeff(point%inSupports.rows(),row);
      }
    }
  }
}

/// Synthesises a bound on the dynamics given a known guard and eigenvectors.
template<class scalar>
void CanonicalSystem<scalar>::synthesiseDynamicBounds(AbstractPolyhedra<scalar> &result,inputType_t inputType,AbstractPolyhedra<scalar> &end)
{
  m_inputType=inputType;
  MatrixS vectors;
  int numVertices;
  getAbstractVertices(end.getDirections(),vectors,numVertices);
  vectors.transposeInPlace();
  MatrixS supports(vectors.rows(),1);
  MatrixS endSupports=end.getSupports();
  int perTemplate=supports.rows()/endSupports.rows();
  for (int i=0;i<endSupports.rows();i++)
  {
    for (int j=0;j<perTemplate;j++) {
      supports.coeffRef(i*perTemplate+j,0)=endSupports.coeff(i,0);
    }
  }
  if (m_inputType>eNoInputs) {
    MatrixS inSupports=m_accelVertices*end.getDirections();
    demergeAccelInSupports(supports,inSupports,endSupports.rows());
  }
  result.load(vectors,supports);
  if (ms_trace_dynamics[eTraceController]) {
    result.logTableau("Full Bounds",true);
  }
  if (isDivergent()) result.removeRedundancies();
  else result.removeRedundancies(1e-4,func::ms_1);
}

/// Retrieves constraints on the controller coefficients based on the Dynamic constraints
template<class scalar>
AbstractPolyhedra<scalar> CanonicalSystem<scalar>::getControllerDynBounds(AbstractPolyhedra<scalar>& reachTube,int &orBlockSize)
{
  //max((I-A-BK)Sv)>max(Rv) -> +p_R+p_S((A-I)^T)<-p_S(BK)
  //(A-I)S-R>-SBK -> -\sum(bji)k_{lj}S_lV_l < (A-I)Sv-B_nR_nv
  MatrixS templates=reachTube.getDirections();

  MatrixS dynamics=m_dynamics;
  for (int i=0;i<m_dimension;i++) dynamics.coeffRef(i,i)-=func::ms_1;
  MatrixS aDirections=dynamics.transpose()*templates;
  MatrixS inDirections=m_sensitivity.transpose()*templates;
  MatrixS aSupports;
  reachTube.maximiseAll(aDirections,aSupports);

  if (!m_reference.isEmpty()) {
    MatrixS inSupports;
    m_reference.maximiseAll(inDirections,inSupports);
    aSupports-=inSupports;
  }

  MatrixS &vertices=reachTube.getVertices();
  orBlockSize=vertices.rows();
  MatrixS supports(aSupports.rows()*vertices.rows(),1);
  MatrixS coefficients=templates.transpose()*m_sensitivity;
  MatrixS faces(supports.rows(),coefficients.cols()*m_dimension);
  for (int row=0;row<templates.cols();row++) {
    for (int i=0;i<orBlockSize;i++) {
      for (int j=0;j<coefficients.cols();j++) {
        faces.block(row*orBlockSize+i,j*m_dimension,1,m_dimension)=coefficients.coeff(row,j)*vertices.row(i);
      }
      supports.coeffRef(row*orBlockSize+i,0)=aSupports.coeff(row,0);
    }
  }
  AbstractPolyhedra<scalar> result(faces.cols());
  result.load(faces,supports);
  return result;
}

/// Retrieves constraints on the controller coefficients based on the I/O constraints
template<class scalar>
AbstractPolyhedra<scalar> CanonicalSystem<scalar>::getControllerInBounds(AbstractPolyhedra<scalar>& reachTube)
{
  MatrixS &vertices=reachTube.getVertices();
  MatrixS directions=m_inputs.getDirections();
  directions.conservativeResize(m_fdimension,directions.cols());
  MatrixS faces(vertices.rows()*directions.cols(),vertices.cols()*m_fdimension);
  MatrixS supports(vertices.rows()*directions.cols(),1);
  MatrixS inputSupports=m_inputs.getSupports();
  for (int row=0;row<vertices.rows();row++) {
    for (int col=0;col<directions.cols();col++) {
      for (int i=0;i<m_fdimension;i++) {
        for (int j=0;j<vertices.cols();j++) {
          faces.coeffRef(row*directions.cols()+col,i*vertices.cols()+j)=vertices.coeff(row,j)*directions.coeff(i,col);
        }
      }
      supports.coeffRef(row*directions.cols()+col)=inputSupports.coeff(col,0);
    }
  }
  AbstractPolyhedra<scalar> result(vertices.cols()*m_fdimension);
  result.load(faces,supports);
  result.removeRedundancies();
  return result;
}

/// Retrieves constraints on the controller coefficients based on stability
template<class scalar>
AbstractPolyhedra<scalar> CanonicalSystem<scalar>::getControllerStabilityBounds(const MatrixS &coefficients,bool nonNegative)
{
  int dimension=coefficients.cols();
  MatrixS faces(2*dimension,dimension);
  faces.block(0,0,dimension,dimension)=MatrixS::Identity(dimension,dimension);
  faces.block(dimension,0,dimension,dimension)=-MatrixS::Identity(dimension,dimension);
  MatrixS supports(2*dimension,1);
  scalar cap=getSpeed();
  for (int row=0;row<dimension-1;row++) {
    supports.coeffRef(row,0)=cap*binomial(dimension,dimension-row-1);
    cap*=getSpeed();
  }
  supports.coeffRef(dimension-1,0)=func::ms_1;
  supports.block(dimension,0,dimension,1)=supports.block(0,0,dimension,1);
  if (nonNegative) {
    for (int row=0;row<dimension;row=2) supports.coeffRef(row,0)=0;
    for (int row=1;row<dimension;row=2) supports.coeffRef(dimension+row,0)=0;
  }
  supports.block(dimension,0,dimension,1)+=coefficients.transpose();
  supports.block(0,0,dimension,1)-=coefficients.transpose();
  AbstractPolyhedra<scalar> result(dimension);
  result.load(faces,supports);
  return result;
}

/// Creates the observer matrices (based on the FWL truncation)
/// and returns the maximum error caused by the truncation
/// for the eigenvalues of the observed system
template <class scalar>
scalar CanonicalSystem<scalar>::makeObserver()
{
  truncate(m_observerDynamics,m_dynamics);
  MatrixS dynamicsError=m_dynamics-m_observerDynamics;
  for (int row=0;row<dynamicsError.rows();row++) { //TODO: should be custom matrix operation
    for (int col=0;col<dynamicsError.cols();col++) {
      if (!func::isPositive(dynamicsError.coeff(row,col))) dynamicsError.coeffRef(row,col)=-func::toLower(dynamicsError.coeff(row,col));
    }
  }
  m_observerDynamicsError=dynamicsError.norm();
  MatrixS colError=dynamicsError.colwise().sum();
  m_controllerGershgorinRadius=func::ms_0;
  for (int col=0;col<colError.cols();col++) {
    if (func::isPositive(colError.coeff(0,col)-m_controllerGershgorinRadius)) m_controllerGershgorinRadius=colError.coeff(0,col);
  }
  MatrixS rowError=dynamicsError.rowwise().sum();
  m_observerGershgorinRadius=func::ms_0;
  for (int row=0;row<rowError.rows();row++) {
    if (func::isPositive(rowError.coeff(row,0)-m_observerGershgorinRadius)) m_controllerGershgorinRadius=rowError.coeff(row,0);
  }

  truncate(m_observerSensitivity,m_sensitivity);
  MatrixS sensitivityError=m_sensitivity-m_observerSensitivity;
  for (int row=0;row<sensitivityError.rows();row++) { //TODO: should be custom matrix operation
    for (int col=0;col<sensitivityError.cols();col++) {
      if (!func::isPositive(sensitivityError.coeff(row,col))) sensitivityError.coeffRef(row,col)=-func::toLower(sensitivityError.coeff(row,col));
    }
  }
  MatrixS colSensitivityError=sensitivityError.colwise().sum();
  m_observerSensitivityError=func::ms_0;
  for (int col=0;col<colSensitivityError.cols();col++) {
    if (func::isPositive(colSensitivityError.coeff(0,col)-m_observerSensitivityError)) m_observerSensitivityError=colSensitivityError.coeff(0,col);
  }
  truncate(m_observerOutputSensitivity,m_outputSensitivity);
  MatrixS outputSensitivityError=m_outputSensitivity-m_observerOutputSensitivity;
  for (int row=0;row<outputSensitivityError.rows();row++) { //TODO: should be custom matrix operation
    for (int col=0;col<outputSensitivityError.cols();col++) {
      if (!func::isPositive(outputSensitivityError.coeff(row,col))) outputSensitivityError.coeffRef(row,col)=-func::toLower(outputSensitivityError.coeff(row,col));
    }
  }
  m_observerOutputSensitivityError=func::ms_0;
  MatrixS rowSensitivityError=outputSensitivityError.rowwise().sum();
  for (int row=0;row<rowSensitivityError.rows();row++) {
    if (func::isPositive(rowSensitivityError.coeff(row,0)-m_observerOutputSensitivityError)) m_observerOutputSensitivityError=rowSensitivityError.coeff(row,0);
  }
  if (ms_trace_dynamics[eTraceController]) {
    ms_logger.logData(m_observerDynamics,"observer dynamics");
    ms_logger.logData(m_observerSensitivity,"observer sensitivity");
    ms_logger.logData(m_observerOutputSensitivity,"observer output sensitivity");
  }
  return m_observerDynamicsError+m_observerSensitivityError+m_observerOutputSensitivityError;
}

/// Creates the observer matrices (based on the FWL truncation)
/// and the closed loop dynamics related to that observer feedback
/// and returns the maximum error caused by the truncation
/// for the eigenvalues of the observed system
template <class scalar>
scalar CanonicalSystem<scalar>::makeObserver(MatrixS &dynamics,MatrixS &sensitivity,bool hasNoise)
{
  scalar result=makeObserver();
  int dimension=2*m_dimension;
  if ((m_observer.rows()!=m_dimension) && (m_observer.cols()!=m_fdimension)) m_observer=MatrixS::Zero(m_dimension,m_fdimension);
  if (&dynamics!=&ms_emptyMatrix) {
    dynamics.conservativeResize(dimension,dimension);
    dynamics.block(0,0,m_dimension,m_dimension)=m_dynamics-m_sensitivity.block(0,0,m_dimension,m_fdimension)*m_feedback;
    dynamics.block(0,m_dimension,m_dimension,m_dimension)=m_observerSensitivity.block(0,0,m_dimension,m_fdimension)*m_feedback;
    dynamics.block(m_dimension,0,m_dimension,m_dimension)=m_dynamics-m_observerDynamics-m_observer*(m_outputSensitivity-m_observerOutputSensitivity);
    dynamics.block(m_dimension,m_dimension,m_dimension,m_dimension)=m_dynamics-m_observer*m_observerOutputSensitivity;
    dynamics.block(0,0,m_dimension,m_dimension)+=(m_sensitivity.block(0,0,m_dimension,m_fdimension)-m_observerSensitivity.block(0,0,m_dimension,m_fdimension))*m_feedback;
    if (ms_trace_dynamics[eTraceController]) {
      MatrixS matrix=m_observerSensitivity*m_feedback;
      ms_logger.logData(matrix,"BK");
      matrix=m_observer*m_observerOutputSensitivity;
      ms_logger.logData(matrix,"LC");
    }
  }
  if (&sensitivity!=&ms_emptyMatrix) {
    int inputs=sensitivity.cols()-(hasNoise ? m_observer.cols() : 0);
    sensitivity.conservativeResize(dimension,sensitivity.cols());
    if (inputs>0) sensitivity.block(m_dimension,0,m_dimension,inputs)=MatrixS::Zero(m_dimension,inputs);
    if (hasNoise) {
      sensitivity.block(0,inputs,m_dimension,m_observer.cols())=MatrixS::Zero(m_dimension,m_observer.cols());
      sensitivity.block(m_dimension,inputs,m_dimension,m_observer.cols())=-m_observer;
    }
  }
  return result;
}

template <class scalar>
bool CanonicalSystem<scalar>::makeClosedSystem(CanonicalSystem<scalar> &closedSystem)
{
  int dimension=(m_inputType>eParametricInputs) ? 4*m_dimension : 2*m_dimension;
  closedSystem.setParams(m_paramValues);
  closedSystem.m_paramValues.coeffRef(eNumStates,0)=dimension;
  for (int i=0;i<3;i++) {
    closedSystem.m_paramValues.coeffRef(eNumInputs,i)=0;
    closedSystem.m_paramValues.coeffRef(eNumVarInputs,i)=0;
  }
  MatrixS dynamics(dimension,dimension);
  AbstractPolyhedra<scalar> guard;
  AbstractPolyhedra<scalar> safeReachTube;
  AbstractPolyhedra<scalar> inputs(m_transformedInputs.getPolyhedra());
  AbstractPolyhedra<scalar> initialState(m_initialState.getPolyhedra());
  if (m_inputType>eParametricInputs) {
    dynamics.block(0,0,dimension,m_dimension)=MatrixS::Zero(dimension,m_dimension);
    dynamics.block(0,m_dimension,m_dimension,m_dimension)=MatrixS::Identity(m_dimension,m_dimension);
    dynamics.block(m_dimension,m_dimension,m_dimension,m_dimension)=m_dynamics;
    dynamics.block(2*m_dimension,m_dimension,2*m_dimension,m_dimension)=MatrixS::Zero(2*m_dimension,m_dimension);
    dynamics.block(0,2*m_dimension,m_dimension,m_dimension)=MatrixS::Identity(m_dimension,m_dimension);
    dynamics.block(m_dimension,2*m_dimension,m_dimension,m_dimension)=MatrixS::Zero(m_dimension,m_dimension);
    dynamics.block(2*m_dimension,2*m_dimension,m_dimension,m_dimension)=MatrixS::Identity(m_dimension,m_dimension);
    dynamics.block(3*m_dimension,2*m_dimension,m_dimension,m_dimension)=MatrixS::Zero(m_dimension,m_dimension);
    dynamics.block(0,3*m_dimension,2*m_dimension,m_dimension)=MatrixS::Zero(2*m_dimension,m_dimension);
    dynamics.block(2*m_dimension,3*m_dimension,m_dimension,m_dimension)=MatrixS::Identity(m_dimension,m_dimension);
    for (int row=0;row<m_dimension;row++) {
      for (int col=0;col<m_dimension;col++) {
        if (func::isPositive(m_dynamics.coeff(row,col))) dynamics.coeffRef(3*m_dimension+row,3*m_dimension+col)=m_dynamics.coeff(row,col);
        else dynamics.coeffRef(3*m_dimension+row,3*m_dimension+col)=-m_dynamics.coeff(row,col);
      }
    }

    m_transformedInputs.getPolyhedra().getSquareAbstraction(inputs,true);
    AbstractPolyhedra<scalar> secondState;
    AbstractPolyhedra<scalar> secondInput;
    initialState.getTransformedPolyhedra(secondState,m_dynamics);
    initialState.concatenate(secondState);
    inputs.getTransformedPolyhedra(secondInput,m_dynamics);
    initialState.concatenate(inputs);
    initialState.concatenate(secondInput);
  }
  else {
    dynamics.block(0,0,m_dimension,m_dimension)=m_dynamics;
    dynamics.block(0,m_dimension,m_dimension,m_dimension)=MatrixS::Zero(m_dimension,m_dimension);
    dynamics.block(m_dimension,0,m_dimension,m_dimension)=MatrixS::Zero(m_dimension,m_dimension);
    dynamics.block(m_dimension,m_dimension,m_dimension,m_dimension)=MatrixS::Identity(m_dimension,m_dimension);
    initialState.concatenate(inputs);
  }
  inputs.clear();
  MatrixS sensitivity=MatrixS(0,0);
  closedSystem.setInputType(eNoInputs);
  closedSystem.changeDimensions(dynamics.rows(),sensitivity.cols(),0,0);
  bool result=closedSystem.load(dynamics,sensitivity,guard,initialState,inputs,safeReachTube);
  return result;
}

/// gets the smallest multiplier that minimises the eigenvalues while remaining in a safe region
template <class scalar>
scalar CanonicalSystem<scalar>::getMinEigenvalueMultiplier(MatrixS &canonical,AbstractPolyhedra<scalar> &bounds)
{
  scalar low=func::ms_0;
  scalar high=func::ms_1;
  int dimension=canonical.cols();
  MatrixS polynomial(1,dimension);
  canonical+=m_polynomialCoefficients;
  bounds.translate(m_polynomialCoefficients);
  for (int k=0;k<10;k++) {
    scalar mid=(high+low)/scalar(2);
    scalar mult=mid;
    for (int i=0;i<dimension;i++) {
      polynomial.coeffRef(0,i)=canonical.coeff(0,i)*mult;
      mult*=mid;
    }
    if (bounds.isInside(polynomial)) {
      high=mid;
      break;
    }
    else {
      low=mid;
    }
  }

  for (int k=0;k<10-func::toInt(func::toUpper(scalar(5)*high));k++) {
    scalar mid=(scalar(3)*high+low)/scalar(4);
    scalar mult=mid;
    for (int i=0;i<dimension;i++) {
      polynomial.coeffRef(0,i)=canonical.coeff(0,i)*mult;
      mult*=mid;
    }
    if (bounds.isInside(polynomial)) high=mid;
    else                             low=mid;
  }
  scalar mult=high;
  for (int i=0;i<dimension;i++) {
    canonical.coeffRef(0,i)*=mult;
    mult*=high;
  }
  canonical-=m_polynomialCoefficients;
  return high;
}

/// Limits the eigenvalues to cap and returns the corresponding polynomial
template <class scalar>
typename CanonicalSystem<scalar>::MatrixS CanonicalSystem<scalar>::capEigenValues(scalar cap, bool split, bool keep)
{
  MatrixC eigenValues=MatrixC::Zero(m_dimension,m_dimension);
  for (int i=0;i<m_dimension;i++) {
    scalar norm=func::normsq(m_eigenValues.coeff(i,i));
    char sign=func::hardSign(norm-cap);
    if (sign<=0) eigenValues.coeffRef(i,i)=m_eigenValues.coeff(i,i);
    else {
      eigenValues.coeffRef(i,i)=m_eigenValues.coeff(i,i)*cap/norm;
    }
  }
  if (keep) m_eigenValues=eigenValues;
  if (split) {
    int dimension=m_dimension/2;
    MatrixC first=eigenValues.block(0,0,dimension,dimension);
    MatrixC second=eigenValues.block(dimension,dimension,dimension,dimension);
    MatrixS controller=getDynamicPolynomialCoefficients(first);
    MatrixS observer=getDynamicPolynomialCoefficients(second);
    MatrixS result(1,m_dimension);
    result.block(0,0,1,dimension)=controller.block(0,1,1,dimension);
    result.block(0,dimension,1,dimension)=observer.block(0,1,1,dimension);
    return result;
  }
  MatrixS result=getDynamicPolynomialCoefficients(eigenValues);
  return result.block(0,1,1,m_dimension);
}

/// Retrives the transform matrix that renders the dynamics in reachable [observable] form
template <class scalar>
void CanonicalSystem<scalar>::getCanonicalTransformMatrices(MatrixS &T, MatrixS &invT, bool observer)
{
  T=m_T;
  invT=m_invT;
  if (observer) {
    int dimension=T.rows();
    T.conservativeResize(dimension+m_W.rows(),dimension+m_W.cols());
    T.block(dimension,dimension,m_W.rows(),m_W.cols())=m_W;
    T.block(0,dimension,dimension,m_W.cols())=MatrixS::Zero(dimension,m_W.cols());
    T.block(dimension,0,m_W.rows(),dimension)=MatrixS::Zero(m_W.rows(),dimension);
    invT.conservativeResize(dimension+m_invW.rows(),dimension+m_invW.cols());
    invT.block(dimension,dimension,m_invW.rows(),m_invW.cols())=m_invW;
    invT.block(0,dimension,dimension,m_invW.cols())=MatrixS::Zero(dimension,m_invW.cols());
    invT.block(dimension,0,m_invW.rows(),dimension)=MatrixS::Zero(m_invW.rows(),dimension);
  }
}

/// Retireves the largest number that need to be represented to execute the model
template <class scalar>
typename CanonicalSystem<scalar>::refScalar CanonicalSystem<scalar>::findMaxValue(bool observer,bool states,bool transforms)
{
  refScalar result=DigitalSystem<scalar>::findMaxValue(observer,states);
  if (transforms) {
    makeReachabilityMatrices(observer);
    result=findMaxValue(m_T,result);
    result=findMaxValue(m_invT,result);
    if (observer) {
      result=findMaxValue(m_W,result);
      result=findMaxValue(m_invW,result);
    }
  }
  return result;
}

/// Updates the feedback and observer matrices only if they have not been set or if they are being forced
template <class scalar>
void CanonicalSystem<scalar>::updateFeedback(const MatrixS &feedback,const MatrixS &observer,bool force)
{
  if (force || (m_feedback.cols()!=feedback.cols()) || (m_feedback.rows()==0)) {
    m_feedback=feedback;
  }
  if (force || (m_observer.rows()!=observer.rows()) || (m_observer.cols()==0)) {
    m_observer=observer;
  }
}

template <class scalar>
int CanonicalSystem<scalar>::getTotalBits(bool observer,int &fracbits)
{
  int totalbits=m_paramValues.coeff(eNumBits,1);
  fracbits=m_paramValues.coeff(eNumBits,2);
  if (totalbits<=0) {
    totalbits=m_paramValues.coeff(eNumBits,0);
    if (totalbits<=0) totalbits=func::getDefaultPrec();
    fracbits=totalbits;
  }
  else if (fracbits==0) {
    refScalar max=findMaxValue(observer);
    max=findMaxValue(getDynamicPolynomialCoefficients(),max);
    int intbits=getBits(max);
    fracbits=totalbits-intbits;
  }
  return totalbits;
}

/// Finds the eigenvalues that correspond to the observer
/// (eigendecomposition does not return ordered eigenvalues)
template <class scalar>
bool CanonicalSystem<scalar>::getObserverRows(scalar coeff,int index,const scalar &observerCoeff,refScalar &error,std::vector<int>& rows,int depth)
{
  index--;
  if (index<0) return false;
  if ((m_conjugatePair[index]>=0) && (depth<2)) {
    return ((index>0) ? getObserverRows(coeff,index,observerCoeff,error,rows,depth) : false);
  }
  int newDepth=depth-1;
  scalar newCoeff=coeff;
  rows[newDepth]=index;
  newCoeff+=m_eigenValues.coeff(index,index).real();
  if (m_conjugatePair[index]>=0) {
    rows[--newDepth]=--index;
    newCoeff+=m_eigenValues.coeff(index,index).real();
  }
  if (newDepth==0) {
    refScalar newError=func::toUpper(abs(newCoeff-observerCoeff));
    if (newError<error) {
      int dimension=m_dimension>>1;
      for (int i=0;i<dimension;i++) rows[i+dimension]=rows[i];
      error=newError;
      if (func::isZero(newError)) return true;
    }
  }
  else if (getObserverRows(newCoeff,index,observerCoeff,error,rows,newDepth)) return true;
  if (index==0) return func::isNearZero(error,.1);
  return ((index>0) ? getObserverRows(coeff,index,observerCoeff,error,rows,depth) : false);
}

/// Rescales the eigenvalues to change the granularity
template<class scalar>
bool CanonicalSystem<scalar>::resample(scalar scale,MatrixS feedbackTarget,MatrixS observerTarget)
{
  if (!DigitalSystem<scalar>::resample(scale)) return false;
  MatrixS coefficients=getDynamicPolynomialCoefficients().block(0,1,1,m_dimension);
  if (ms_trace_dynamics[eTraceRefinements]) {
    ms_logger.logData(coefficients,"resampled coefficients");
    ms_logger.logData(feedbackTarget,"feedback target");
    ms_logger.logData(observerTarget,"observer target");
  }
  if (feedbackTarget.cols()==m_dimension) feedbackTarget-=coefficients;
  if (observerTarget.cols()==m_dimension) observerTarget-=coefficients;
  makeReachabilityMatrices(m_observer.rows()>0);
  if (feedbackTarget.cols()==m_dimension) {
    m_feedback=feedbackTarget*m_T;
    if (ms_trace_dynamics[eTraceRefinements]) ms_logger.logData(m_feedback,"resampled feedback");
  }
  if (observerTarget.cols()==m_dimension) {
    m_observer=m_invW*observerTarget.transpose();
    if (ms_trace_dynamics[eTraceRefinements]) ms_logger.logData(m_observer,"resampled observer");
  }
  return true;
}

#ifdef USE_LDOUBLE
  #ifdef USE_SINGLES
    template class CanonicalSystem<double_type>;
  #endif
  #ifdef USE_INTERVALS
    template class CanonicalSystem<ldinterval>;
  #endif
#endif
#ifdef USE_MPREAL
  #ifdef USE_SINGLES
    template class CanonicalSystem<mpfr::mpreal>;
  #endif
  #ifdef USE_INTERVALS
    template class CanonicalSystem<mpinterval>;
  #endif
#endif

}
