//Authors: Dario Cattaruzza, Alessandro Abate, Peter Schrammel, Daniel Kroening
//University of Oxford 2016
//This code is supplied under the BSD license agreement (see license.txt)
#include "DigitalSystem.h"
#include <boost/timer.hpp>
#include <set>
#include <Eigen/QR>
#include <Eigen/Eigenvalues>

namespace abstract{

template <class scalar>
bvType_t DigitalSystem<scalar>::ms_bvType=eFloatBV;

/// Constructs an empty buffer
template<class scalar>
DigitalSystem<scalar>::DigitalSystem(int dimension,int idimension,int odimension,bool trace) :
  DynamicalSystem<scalar>(dimension,idimension,odimension,trace),
  m_eigenValueCap(func::ms_1),
  m_eigenValueMultiplierMin(func::ms_1),
  m_eigenValueMultiplierMax(func::ms_1)
{}

/// Copies an existing object
template<class scalar>
void DigitalSystem<scalar>::copy(const DigitalSystem &source)
{
  DynamicalSystem<scalar>::copy(source);
  m_eigenValueCap=source.m_eigenValueCap;
  m_samplingCap=source.m_samplingCap;
  m_eigenValueMultiplierMin=source.m_eigenValueMultiplierMin;
  m_eigenValueMultiplierMax=source.m_eigenValueMultiplierMax;
}

/// retrieves the Jordan form of the system sampled at SampleTime
template<class scalar>
typename DigitalSystem<scalar>::MatrixC DigitalSystem<scalar>::getSampledJordanForm(scalar sampleTime)
{
  MatrixC eigenValues=m_eigenValues;
  for (int i=0;i<m_dimension;i++) {
    complexS power=eigenValues.coeffRef(i,i)*sampleTime;
    scalar mag=exp(power.real());
    scalar real=mag*func::cosine(power.imag());
    scalar imag=mag*func::sine(power.imag());
    complexS value=complexS(real,imag);
    eigenValues.coeffRef(i,i)=value;
    powerS div=1;
    for (int j=1;j<m_jordanIndex[i];j++) {
      div*=j;
      eigenValues.coeffRef(i-j,i)=eigenValues.coeffRef(i,i)*func::pow(sampleTime,j)/scalar(div);
    }
  }
  return eigenValues;
}

/// retrieves the dynamics of the system sampled at SampleTime
template<class scalar>
typename DigitalSystem<scalar>::MatrixS DigitalSystem<scalar>::getSampledDynamics(scalar sampleTime)
{
  MatrixS pseudoEigenValues=jordanToPseudoJordan(getSampledJordanForm(sampleTime),this->eToEigenValues);
  return m_pseudoEigenVectors*pseudoEigenValues*m_invPseudoEigenVectors;
}

/// retrieves the delay noise matrix
template<class scalar>
typename DigitalSystem<scalar>::MatrixS DigitalSystem<scalar>::getDelayNoise(scalar sampleTime, scalar maxDelay)
{
  MatrixS discreteDynamics=getSampledDynamics(sampleTime);
  MatrixS delayDynamics=getSampledDynamics(maxDelay);
  MatrixS finalDynamics=getSampledDynamics(sampleTime-maxDelay);
  MatrixS reverseDynamics=makeInverse(m_dynamics);
  if (m_outputSensitivity.rows()*m_outputSensitivity.cols()<=0) {
    m_outputSensitivity=MatrixS::Identity(m_dimension,m_dimension);
  }
  MatrixS vectors(2*m_outputSensitivity.rows(),m_sensitivity.cols());
  vectors.block(0,0,m_outputSensitivity.rows(),m_sensitivity.cols())=m_outputSensitivity*reverseDynamics*(discreteDynamics-finalDynamics)*m_sensitivity;
  vectors.block(m_outputSensitivity.rows(),0,m_outputSensitivity.rows(),m_sensitivity.cols())=m_outputSensitivity*finalDynamics*reverseDynamics*(delayDynamics-MatrixS::Identity(m_dimension,m_dimension))*m_sensitivity;
  vectors.transposeInPlace();
  MatrixS result;
  m_inputs.maximiseAll(vectors,result);
  return result;
}

/// turns a continuous time system into a discrete
template<class scalar>
bool DigitalSystem<scalar>::sample(scalar sampleTime,scalar delayTime, bool strict)
{
  if (ms_trace_dynamics[eTraceSampling]) ms_logger.logData(sampleTime,"Sampling",true);
  m_sampleTime=sampleTime;
  m_delayTime=delayTime;
  if (m_cosFactors.rows()<=m_dimension) m_cosFactors.conservativeResize(m_dimension+1,1);
  m_cosFactors.coeffRef(m_dimension,0)=func::ms_1;
  scalar sum=func::ms_0;
  for (int i=0;i<m_dimension;i++) {
    complexS power=m_eigenValues.coeffRef(i,i)*sampleTime;
    scalar mag=exp(power.real());
    if (strict && func::isNearZero(mag,1e-6)) return false;
    sum+=mag;
    scalar real=mag*func::cosine(power.imag());
    scalar imag=mag*func::sine(power.imag());
    complexS value=complexS(real,imag);
    m_eigenValues.coeffRef(i,i)=value;
    m_eigenNorms.coeffRef(i,0)=func::norm2(value);
    if (m_conjugatePair[i]>=0) {
      m_cosFactors.coeffRef(i,0)=m_eigenNorms.coeffRef(i,0)/value.real();
      m_cosFactors.coeffRef(m_dimension,0)*=m_cosFactors.coeffRef(i,0);
    }
    powerS div=1;
    for (int j=1;j<m_jordanIndex[i];j++) {
      div*=j;
      m_eigenValues.coeffRef(i-j,i)=m_eigenValues.coeffRef(i,i)*func::pow(sampleTime,j)/scalar(div);
    }
  }
  m_pseudoEigenValues=jordanToPseudoJordan(m_eigenValues,this->eToEigenValues);
  MatrixS discrete_dynamics=m_pseudoEigenVectors*m_pseudoEigenValues*m_invPseudoEigenVectors;
  MatrixS integrator=discrete_dynamics-MatrixS::Identity(m_dimension,m_dimension);
  integrator=makeInverse(m_dynamics)*integrator;
  if (ms_trace_dynamics[eTraceSampling]) {
    ms_logger.logData(m_dynamics,"Dynamics");
    ms_logger.logData(m_sensitivity,"Sensitivity");
    ms_logger.logData(discrete_dynamics,"Discrete Dynamics");
    ms_logger.logData(integrator,"Integrator");
    ms_logger.logData(m_pseudoEigenValues,"Discrete PseudoEigenValues");
    ms_logger.logData(m_pseudoEigenVectors,"Discrete PseudoEigenVectors");
    ms_logger.logData(m_invPseudoEigenVectors,"Discrete InvPseudoEigenVectors");
  }
  m_dynamics=discrete_dynamics;
  m_sensitivity=integrator*m_sensitivity;
  processInputs(true);
  calculateBlockSVD();
  calculateInvIminJ();
  calculateF();
  calculateInvIminF();
  if (ms_trace_dynamics[eTraceDynamics]) {
    ms_logger.logData(m_pseudoInvIminJ,"InvIminJ:");
    ms_logger.logData(m_invIminF,"InvIminF:");
  }
  findFrequencies();
  findZeniths();
  if (ms_trace_dynamics[eTraceSampling]) {
    ms_logger.logData(m_sensitivity,"Discrete Sensitivity");
  }
  return !func::isNearZero(sum,1e-6);
}

/// turns a discrete time system into a continuous one
template<class scalar>
void DigitalSystem<scalar>::unsample(scalar sampleTime)
{
  m_sampleTime=sampleTime;
  /*if (m_cosFactors.rows()<=m_dimension) m_cosFactors.conservativeResize(m_dimension+1,1);
  m_cosFactors.coeffRef(m_dimension,0)=func::ms_1;*/
  for (int i=0;i<m_dimension;i++) {
    complexS coef=m_eigenValues.coeffRef(i,i);
    scalar mag=log(func::norm2(coef));
    scalar angle=func::invtan(coef.imag(),coef.real());
    complexS value=complexS(mag,angle)/sampleTime;
    m_eigenValues.coeffRef(i,i)=value;
    m_eigenNorms.coeffRef(i,0)=func::norm2(value);
    /*if (m_conjugatePair[i]>=0) {
      m_cosFactors.coeffRef(i,0)=m_eigenNorms.coeffRef(i,0)/value.real();
      m_cosFactors.coeffRef(m_dimension,0)*=m_cosFactors.coeffRef(i,0);
    }
    powerS div=1;
    for (int j=1;j<m_jordanIndex[i];j++) {
      div*=j;
      m_eigenValues.coeffRef(i-j,i)=m_eigenValues.coeffRef(i,i)*func::pow(sampleTime,j)/scalar(div);
    }*/
  }
  m_pseudoEigenValues=jordanToPseudoJordan(m_eigenValues,this->eToEigenValues);
  MatrixS continuous_dynamics=m_pseudoEigenVectors*m_pseudoEigenValues*m_invPseudoEigenVectors;
  MatrixS derivator=m_dynamics-MatrixS::Identity(m_dimension,m_dimension);
  derivator=makeInverse(derivator)*continuous_dynamics;
  if (ms_trace_dynamics[eTraceSampling]) {
    ms_logger.logData(m_dynamics,"Discrete Dynamics");
    ms_logger.logData(m_sensitivity,"Discrete Sensitivity");
    ms_logger.logData(continuous_dynamics,"Dynamics");
    ms_logger.logData(derivator,"Derivator");
    ms_logger.logData(m_pseudoEigenValues,"PseudoEigenValues");
    ms_logger.logData(m_pseudoEigenVectors,"PseudoEigenVectors");
    ms_logger.logData(m_invPseudoEigenVectors,"InvPseudoEigenVectors");
  }
  m_sensitivity=derivator*m_sensitivity;
  m_dynamics=continuous_dynamics;
  if (ms_trace_dynamics[eTraceSampling])
  {
    ms_logger.logData(m_sensitivity,"Continuous Sensitivity");
  }
  sample(sampleTime,0);
}

template<class scalar>
bool DigitalSystem<scalar>::sample(std::string &sampleTime)
{
  scalar time,delay=0;
  char* pEnd=func::toScalar(sampleTime.data(),time);
  if (pEnd==sampleTime.data()) return false;
  if (pEnd[0]==':') {
    pEnd++;
    char *pStart=pEnd;
    if (pEnd[0]!=':') pEnd=func::toScalar(pEnd,delay);
    if (pEnd==pStart) return false;
  }
  char sign=func::hardSign(time);
  m_samplingCap=time;
  if ((sign>0) && (pEnd[0]==':')) {
    pEnd++;
    scalar max,min=time;
    if (pEnd==func::toScalar(pEnd,max)) return false;
    MatrixC eigenValues=m_eigenValues;
    MatrixS sensitivity=m_sensitivity;
    sample(min,delay);
    if (func::isNegative(max-min)) {
      m_samplingCap=max;
      if (ms_fullAnswers) {
        ms_logger.logData(m_sampleTime,"Sampled",true);
        ms_logger.logData(getDescription());
      }
      return true;
    }
    refScalar minCap=findMaxValue(true,false,true);
    if (ms_trace_dynamics[eTraceSampling]) ms_logger.logData(minCap,"minValue",true);
    m_eigenValues=eigenValues;
    m_sensitivity=sensitivity;
    while (!sample(max,delay) && func::isPositive(max-min)) {
      if (ms_trace_dynamics[eTraceSampling]) {
        ms_logger.logData(max,"sampling time too long",true);
      }
      m_eigenValues=eigenValues;
      m_sensitivity=sensitivity;
      max*=scalar(0.5);
    }
    refScalar maxCap=findMaxValue(true,false,true);
    if (ms_trace_dynamics[eTraceSampling]) ms_logger.logData(maxCap,"maxValue",true);
    scalar m3=3;
    scalar d4=4;
    scalar m100=100;
    while(func::isPositive(max-min-delay*m100)) {
      scalar midmin=(m3*min+max)/d4;
      m_eigenValues=eigenValues;
      m_sensitivity=sensitivity;
      sample(midmin,delay);
      refScalar midminCap=findMaxValue(true,false,true);
      scalar midmax=(min+m3*max)/d4;
      m_eigenValues=eigenValues;
      m_sensitivity=sensitivity;
      sample(midmax,delay);
      refScalar midmaxCap=findMaxValue(true,false,true);
      if (minCap<midminCap) {
        maxCap=midminCap;
        max=midmin;
      }
      else if (maxCap<midmaxCap) {
        minCap=midmaxCap;
        min=midmax;
      }
      else if (midminCap<midmaxCap) {
        maxCap=midmaxCap;
        max=midmax;
        if (minCap>midmaxCap) {
          minCap=midminCap;
          min=midmin;
        }
      }
      else {
        minCap=midminCap;
        min=midmin;
        if (maxCap>midminCap) {
          maxCap=midmaxCap;
          max=midmax;
        }
      }
    }
    m_eigenValues=eigenValues;
    m_sensitivity=sensitivity;
    time=(max+min)/scalar(2);
  }
  if (sign>0)       sample(time,delay);
  else if (sign<0)  unsample(-time);
  else              return false;
  if (ms_fullAnswers) {
    ms_logger.logData(m_sampleTime,"Sampled",true);
    ms_logger.logData(getDescription());
  }
  return true;
}

/// Rescales the eigenvalues to change the granularity
template<class scalar>
bool DigitalSystem<scalar>::resample(scalar scale)
{
  if (m_hasMultiplicities) return false;
  m_sampleTime*=scale;
  ms_logger.logData(m_sampleTime,"resampling",true);
  if (m_cosFactors.rows()<=m_dimension) m_cosFactors.conservativeResize(m_dimension+1,1);
  m_cosFactors.coeffRef(m_dimension,0)=func::ms_1;
  m_eigenValueCap=exp(log(m_eigenValueCap)*scale);
  for(int i=0;i<m_dimension;i++) {
    scalar mag=exp(log(func::norm2(m_eigenValues.coeffRef(i,i)))*scale);
    scalar arg=func::invtan(m_eigenValues.coeffRef(i,i).imag(),m_eigenValues.coeffRef(i,i).real());
    if (func::isPositive(arg-func::ms_pi)) arg-=func::ms_2_pi;
    arg*=scale;
    complexS value(mag*func::cosine(arg),func::sine(arg));
    m_eigenValues.coeffRef(i,i)=value;
    m_eigenNorms.coeffRef(i,0)=func::norm2(m_eigenValues.coeffRef(i,i));
    if (m_conjugatePair[i]>=0) {
      m_cosFactors.coeffRef(i,0)=m_eigenNorms.coeffRef(i,0)/m_eigenValues.coeffRef(i,i).real();
      m_cosFactors.coeffRef(m_dimension,0)*=m_cosFactors.coeffRef(i,0);
    }
  }
  if (ms_trace_dynamics[eTraceEigen]) ms_logger.logData(m_eigenValues,"resampled eigenvalues");
  m_pseudoEigenValues=jordanToPseudoJordan(m_eigenValues,this->eToEigenValues);
  if (ms_trace_dynamics[eTracePseudo]) ms_logger.logData(m_pseudoEigenValues,"resampled pseudoEigenvalues");
  MatrixS dynamics=m_pseudoEigenVectors*m_pseudoEigenValues*m_invPseudoEigenVectors;
  MatrixS integrator=(MatrixS::Identity(m_dimension,m_dimension)-dynamics)*(MatrixS::Identity(m_dimension,m_dimension)-m_dynamics).inverse();
  m_sensitivity=integrator*m_sensitivity;
  m_dynamics=dynamics;
  if (ms_trace_dynamics[eTraceDynamics]) ms_logger.logData(m_dynamics,"resampled dynamics");
  findFrequencies();
  findZeniths();
  return true;
}

template<class scalar>
bool DigitalSystem<scalar>::setSpeed(std::string &max)
{
  scalar speed;
  if (func::toScalar(max.data(),speed)==max.data()) return false;
  m_eigenValueCap=speed;
  return true;
}

/// Creates a model for the quantization noise as an input specification
template<class scalar>
AbstractPolyhedra<scalar> DigitalSystem<scalar>::generateNoiseInput(scalar sampling,scalar delay)
{
  int odimension=(m_odimension>0) ? m_odimension : m_dimension;
  int ndimension=1;
  AbstractPolyhedra<scalar> result(ndimension);
  int fbits;
  fbits=m_paramValues.coeff(eNumBits,2);
  if (fbits<=0) fbits=m_paramValues.coeff(eNumBits,0);
  if (fbits<=0) fbits=func::getDefaultPrec();
  scalar lsb=func::pow(this->ms_two,-fbits);
  MatrixS faces(2*ndimension,ndimension);
  MatrixS supports(2*ndimension,1);
  faces.block(0,0,ndimension,ndimension)=MatrixS::Identity(ndimension,ndimension);
  faces.block(ndimension,0,ndimension,ndimension)=-MatrixS::Identity(ndimension,ndimension);
  scalar fb_noise=m_feedback.cwiseAbs().sum()/*lpNorm<1>()*/+func::ms_1;//|Noise|<(|K|_1+1)lsb assuming q1,q2=1lsb,(q3:=q1->K) = (|K|_1q1+q2)
  supports.coeffRef(0,0)=lsb*(fb_noise);
  if (!func::isZero(delay)) supports.coeffRef(0,0)+=getDelayNoise(sampling,delay).norm();
  supports.coeffRef(1,0)=supports.coeff(0,0);
  result.load(faces,supports);
  return result;
}

/// Creates a model for the input of the closed loop
template<class scalar>
AbstractPolyhedra<scalar> DigitalSystem<scalar>::generateFeedbackInput(int fdimension,bool makeNoise,MatrixS &sensitivity,scalar sampling,scalar delay)
{
  AbstractPolyhedra<scalar> inputs(0);
  if (m_reference.getDimension()>0) {
    inputs.copy(m_reference);
    fdimension=m_reference.getDimension();
  }
  if (m_idimension>fdimension) {
    MatrixS inputFaces=m_inputs.getFaceDirections().rightCols(m_idimension-fdimension);
    MatrixS inSupports=m_inputs.getSupports();
    inputs.concatenate(inputFaces,inSupports);
  }
  if (makeNoise) {
    //m_closedLoop.setInputType(eVariableInputs);
    AbstractPolyhedra<scalar> noiseInputs=generateNoiseInput(sampling,delay);
    inputs.concatenate(noiseInputs);
    sensitivity.conservativeResize(m_dimension,inputs.getDimension());
    sensitivity.block(0,inputs.getDimension()-noiseInputs.getDimension(),m_dimension,noiseInputs.getDimension())=MatrixS::Ones(m_dimension,noiseInputs.getDimension());
  }
  return inputs;
}

/// Truncates the value to the current controller precision
template <class scalar>
scalar DigitalSystem<scalar>::truncateCoefficient(const scalar& coeff)
{
  int totalbits=m_paramValues.coeff(eNumBits,1);
  long long fracbits=m_paramValues.coeff(eNumBits,2);
  if (totalbits<=0) {
    totalbits=m_paramValues.coeff(eNumBits,0);
    if (totalbits<=0) totalbits=func::getDefaultPrec();
    fracbits=totalbits;
  }
  else if (fracbits==0) fracbits=totalbits>>1;
  refScalar fxp_one = (1 << fracbits);
  if (ms_bvType==eFixedBV) {
    int intbits=totalbits-fracbits;
    scalar lower=floor(func::toLower(coeff) * fxp_one)/fxp_one;
    scalar upper=floor(func::toUpper(coeff) * fxp_one)/fxp_one;
    scalar new_coeff=func::getHull(lower,upper);
    scalar fxp_max=(1 << intbits);
    return (func::isPositive(new_coeff-fxp_max)) ? fxp_max : new_coeff;
  }
  int expMask=(1<<(totalbits-fracbits-1))-1;
  fxp_one*=2;
  typename func::exponent exp;
  scalar upper=floor(func::extractExponent(func::toUpper(coeff),exp)*fxp_one);
  if (exp>expMask) upper=(1<<(expMask+1))-(1.0/(1<<expMask));
  else if (exp>=0) upper*=scalar(1<<exp);
  else if (-exp>expMask) upper=0;
  else upper/=scalar(1<<-exp);
  scalar lower=floor(func::extractExponent(func::toLower(coeff),exp)*fxp_one);
  if (exp>expMask) upper=(1<<(expMask+1))-(1.0/(1<<expMask));
  else if (exp>=0) lower*=scalar(1<<exp);
  else if (-exp>expMask) lower=0;
  else lower/=scalar(1<<-exp);
  scalar newCoeff=func::getHull(lower,upper);
  newCoeff/=fxp_one;
  return newCoeff;
}

/// Truncates the values of src to the current controller precision
template <class scalar>
void DigitalSystem<scalar>::truncate(MatrixS &dest,const MatrixS &src)
{
  dest.conservativeResize(src.rows(),src.cols());
  for (int row=0;row<src.rows();row++) {
    for (int col=0;col<src.cols();col++) {
      dest.coeffRef(row,col)=truncateCoefficient(src.coeff(row,col));
    }
  }
}

/// Retireves the number of bits needed to represent max
template <class scalar>
int DigitalSystem<scalar>::getBits(refScalar max)
{
  // Retrieves the integer bits (fixed) or exponent bits (float)
  // required to represent the range of max
  int bits=1;// Minimum 1
  if (ms_bvType==eFixedBV) {
    long long maxNumber=func::toInt(max);
    while((maxNumber>>=1)!=0) bits++;
  }
  else {
    typename func::exponent exp;
    func::extractExponent(max,exp);
    if (exp<0) exp=-exp;
    while((exp>>=1)!=0) bits++;
  }
  bits++;// Sign bit;
  return bits;
}

/// Retireves the number of bits needed to represent the vertices of the polyhedra
template <class scalar>
int DigitalSystem<scalar>::getBits(AbstractPolyhedra<scalar> polyhedra)
{
  MatrixS box=polyhedra.boundingHyperBox();
  return getBits(findMaxValue(box));
}

/// Clips the matrix to fit the FWL given by bits
template <class scalar>
void DigitalSystem<scalar>::fitToSpec(MatrixS &matrix, int totalbits,int fracbits)
{
  if (ms_bvType==eFixedBV) {
    int intbits=totalbits-fracbits;
    refScalar max=pow(refScalar(2),intbits-1);
    refScalar min=-max;
    refScalar one=pow(refScalar(2),fracbits);
    for (int row=0;row<matrix.rows();row++) {
      for (int col=0;col<matrix.cols();col++) {
        matrix.coeffRef(row,col)=floor(func::toCentre(matrix.coeff(row,col)*one))/one;
        if (func::toUpper(matrix.coeff(row,col))>max) matrix.coeffRef(row,col)=max;
        if (func::toLower(matrix.coeff(row,col))<min) matrix.coeffRef(row,col)=min;
      }
    }
  }
  else {
    int expbits=totalbits-fracbits;
    refScalar one=pow(refScalar(2),fracbits);
    int intbits=2<<(expbits-1);
    refScalar max=pow(refScalar(2),intbits);
    refScalar min=-max;
    typename func::exponent exp;
    for (int row=0;row<matrix.rows();row++) {
      for (int col=0;col<matrix.cols();col++) {
        matrix.coeffRef(row,col)=floor(func::extractExponent(func::toCentre(matrix.coeff(row,col)), exp)*one)/one;
        matrix.coeffRef(row,col)*=pow(refScalar(2),exp);
        if (func::toUpper(matrix.coeff(row,col))>max) matrix.coeffRef(row,col)=max;
        if (func::toLower(matrix.coeff(row,col))<min) matrix.coeffRef(row,col)=min;
      }
    }

  }
}

/// Normalizes the statespace
template <class scalar>
bool DigitalSystem<scalar>::scaleSystem(scalar scale)
{
  if (func::isZero(scale-func::ms_1)) {
    MatrixS safe=m_safeReachTube.getPolyhedra().boundingHyperBox();
    if (safe.rows()*safe.cols()<=0) return false;
    scale/=findMaxValue(safe);
  }
  m_safeReachTube.getPolyhedra().transform(scale);
  m_safeReachTube.clear(false);
  m_initialState.getPolyhedra().transform(scale);
  m_initialState.clear(false);
  m_inputs.transform(scale);
  m_transformedInputs.getPolyhedra().transform(scale);
  m_transformedInputs.clear(false);
  m_reference.transform(scale);
  m_outputs.transform(scale);
  return true;
}

/// Retireves the largest number that need to represent the matrix
template <class scalar>
typename DigitalSystem<scalar>::refScalar DigitalSystem<scalar>::findMaxValue(const MatrixS &matrix,refScalar result,refScalar scale)
{
  refScalar max=0;
  for (int row=0;row<matrix.rows();row++) {
    for (int col=0;col<matrix.cols();col++) {
      if (func::toUpper(matrix.coeff(row,col))>max) max=func::toUpper(matrix.coeff(row,col));
      if (-func::toLower(matrix.coeff(row,col))>max) max=-func::toLower(matrix.coeff(row,col));
    }
  }
  if (!func::isZero(scale)) max/=scale;
  if (max>result) result=max;
  return result;
}

/// Retireves the largest difference between any two non-zero numbers in a matrix
template <class scalar>
scalar DigitalSystem<scalar>::findRange(const MatrixS &matrix)
{
  refScalar max=findMaxValue(matrix);
  refScalar min=max;
  for (int row=0;row<matrix.rows();row++) {
    for (int col=0;col<matrix.cols();col++) {
      char sign=func::hardSign(matrix.coeff(row,col));
      if (sign>0) {
        if (func::toLower(matrix.coeff(row,col))<min) min=func::toLower(matrix.coeff(row,col));
      }
      else if (sign<0) {
        if (-func::toUpper(matrix.coeff(row,col))<min) min=-func::toUpper(matrix.coeff(row,col));
      }
    }
  }
  return scalar(max)/min;
}

/// Retireves the best scale factor usable to reach target bits
template <class scalar>
scalar DigitalSystem<scalar>::findScale(MatrixS &matrix,int targetBits,bool autoScale)
{
  refScalar max=findMaxValue(matrix);
  scalar min=max/findRange(matrix);
  int rangeT=func::toInt(func::toUpper(min));
  int scaleT=1;
  int bits=getBits(max);
  if (bits>targetBits) {
    int range=1<<(bits-targetBits);
    if (rangeT>range) {
      if (autoScale) matrix/=scaleT;
      return range;
    }
    while(rangeT>1) {
      scaleT<<=1;
      rangeT>>=1;
    }
    if (autoScale) matrix/=scaleT;
  }
  return scaleT;
}


/// Retireves the largest number that need to be represented to execute the model
template <class scalar>
typename DigitalSystem<scalar>::refScalar DigitalSystem<scalar>::findMaxValue(bool dynamics,bool states,bool intermediate)
{
  refScalar result=func::ms_0;
  if (dynamics) {
    result=findMaxValue(m_dynamics,result);
    result=findMaxValue(m_sensitivity,result);
    result=findMaxValue(m_outputSensitivity,result);
  }
  if (states) {
    if (m_outputGuard.isEmpty() && m_safeReachTube.getPolyhedra().isEmpty()) {
      MatrixS init=m_initialState.getPolyhedra().boundingHyperBox();
      result=findMaxValue(init,result);
    }
    MatrixS inputs=m_inputs.boundingHyperBox();
    result=findMaxValue(inputs,result);
    MatrixS safe=m_outputGuard.isEmpty() ? m_safeReachTube.getPolyhedra().boundingHyperBox() : m_outputGuard.boundingHyperBox();
    result=findMaxValue(safe,result);
  }
  return result;
}

#ifdef USE_LDOUBLE
  #ifdef USE_SINGLES
    template class DigitalSystem<double_type>;
  #endif
  #ifdef USE_INTERVALS
    template class DigitalSystem<ldinterval>;
  #endif
#endif
#ifdef USE_MPREAL
  #ifdef USE_SINGLES
    template class DigitalSystem<mpfr::mpreal>;
  #endif
  #ifdef USE_INTERVALS
    template class DigitalSystem<mpinterval>;
  #endif
#endif

}
