//Authors: Dario Cattaruzza, Alessandro Abate, Peter Schrammel, Daniel Kroening
//University of Oxford 2016
//This code is supplied under the BSD license agreement (see license.txt)

#include "SparseRowSort.h"
#include <map>
#include <set>
#include <math.h>
#include <boost/timer.hpp>
#include "MatrixToString.h"

namespace abstract {
 /*  quicksorts matrix rows based on the column order (first column is most significant)
 */
template <class scalar>
SparseSortedMatrix<scalar>::SparseSortedMatrix(int rows,int cols) : sparseMatrix(rows,cols) {}

template <class scalar>
SparseSortedMatrix<scalar>::SparseSortedMatrix(sparseMatrix &source,const OrderType &type,bool transpose,bool norm) : sparseMatrix(transpose ? source.cols() : source.rows(),transpose ? source.rows() : source.cols())
{
  if (transpose) {
    for (int row=0;row<source.rows();row++) {
      sparseVector &data=source[row];
      for (typename sparseVector::iterator it=data.begin();it!=data.end();it++) {
        (*this)[it->first][row]=it->second;
      }
    }
  }
  else {
    static_cast<sparseMatrix>(*this)=source;
    this->m_cols=source.cols();
  }
  if (type==LexCos) normalize();
  if (type!=LexNone) ComputeRowOrderVector(type);
}

template <class scalar>
char SparseSortedMatrix<scalar>::compareZeroOrderRows(int row1,int row2)
{
  int trueRow1=zeroOrder(row1);
  int trueRow2=zeroOrder(row2);
  for (int col=0;col<cols();col++) {
    char sign=func::hardSign(coeff(trueRow1,col)-coeff(trueRow2,col));
    if (sign!=0) return sign;
  }
  return 0;
}

template <class scalar>
scalar SparseSortedMatrix<scalar>::diffZeroOrderRows(int row1,int row2)
{
  scalar result=func::ms_0;
  int trueRow1=zeroOrder(row1);
  int trueRow2=zeroOrder(row2);
  for (int col=0;col<cols();col++) {
    result+=abs(coeff(trueRow1,col)-coeff(trueRow2,col));
  }
  return result;
}

template <class scalar>
scalar SparseSortedMatrix<scalar>::diffZeroOrderRows(int row1,int row2,MatrixS &ranges)
{
  scalar result=func::ms_0;
  int trueRow1=zeroOrder(row1);
  int trueRow2=zeroOrder(row2);
  for (int col=0;col<cols();col++) {
    result+=func::toWidth(coeff(trueRow1,col)-coeff(trueRow2,col))*ranges.coeff(0,col);
  }
  return result;
}

template <class scalar>
bool SparseSortedMatrix<scalar>::SmallerCos(std::vector<scalar> &cosines, const int row1,const int row2)
{
  scalar result=cosines[row1]-cosines[row2];
  char sign=func::hardSign(result);
  if (sign!=0) return (sign<0);
  return false;
}

template <class scalar>
bool SparseSortedMatrix<scalar>::LargerCos(std::vector<scalar> &cosines, const int row1,const int row2)
{
  scalar result=cosines[row1]-cosines[row2];
  char sign=func::hardSign(result);
  if (sign!=0) return (sign > 0);
  return false;
}

template <class scalar>
long SparseSortedMatrix<scalar>::subCosPartition(std::vector<int> &order, std::vector<scalar> &cosines, const long p, const long r)
{
  long i,j,tmp;
  i=p-1;
  j=r+1;
  while (true){
    while (LargerCos(cosines,order[--j]-1,order[p]-1));
    while (SmallerCos(cosines,order[++i]-1,order[p]-1));
    if (i<j){
      tmp=order[i];
      order[i]=order[j];
      order[j]=tmp;
    }
    else return j;
  }
}

template <class scalar>
void SparseSortedMatrix<scalar>::QuickCosSort(std::vector<int> &order, std::vector<scalar> &cosines, const long p, const long r)
{
  if (p < r){
    long q = subCosPartition(order, cosines, p, r);
    QuickCosSort(order, cosines, p, q);
    QuickCosSort(order, cosines, q+1, r);
  }
}

template <class scalar>
void SparseSortedMatrix<scalar>::cosSort(std::vector<int> &order, std::vector<scalar> &cosines)
{
  int numRows=rows()+1;
  for(int i=0; i<=numRows; i++) order[i]=i;
  for (int row=1;row<rows();row++) {
    scalar value=coeff(0,0)*coeff(row,0);
    for (int col=1;col<cols();col++) func::madd(value,coeff(0,col),coeff(row,col));
    cosines[row]=value;
  }
  QuickCosSort(order, cosines, 1, rows());
  for(int i=1; i<=rows()/2;i++){   /* reverse the order */
    int temp=order[i];
    order[i]=order[numRows-i];
    order[numRows-i]=temp;
  }
}

template <class scalar>
bool SparseSortedMatrix<scalar>::SmallerAngleRow(const int row1,const int row2)
{
  for (int j=0;j<cols()-1;j++) {
    scalar angle1=coeff(row1,j)/coeff(row1,j+1);
    angle1=atan(angle1);
    if (func::isNegative(coeff(row1,j))) angle1=func::const_pi(func::ms_1)-angle1;
    if (func::isNegative(coeff(row1,j+1))) angle1+=func::const_pi(func::ms_1);
    scalar angle2=coeff(row2,j)/coeff(row2,j+1);
    angle2=atan(angle2);
    if (func::isNegative(coeff(row2,j))) angle2=func::const_pi(func::ms_1)-angle2;
    if (func::isNegative(coeff(row2,j+1))) angle2+=func::const_pi(func::ms_1);
    angle1-=angle2;
    char sign=func::hardSign(angle1);
    if (sign!=0) return (sign<0);
  }
  return false;
}

template <class scalar>
bool SparseSortedMatrix<scalar>::LargerAngleRow(const int row1,const int row2)
{
  for (int j=0;j<cols()-1;j++) {
    scalar angle1=coeff(row1,j)/coeff(row1,j+1);
    angle1=atan(angle1);
    if (coeff(row1,j)<func::ms_0) angle1+=func::const_pi(func::ms_1);
    scalar angle2=coeff(row2,j)/coeff(row2,j+1);
    angle2=atan(angle2);
    if (func::isNegative(coeff(row2,j))) angle2+=func::const_pi(func::ms_1);
    angle1-=angle2;
    char sign=func::hardSign(angle1);
    if (sign!=0) return (sign>0);
  }
  return false;
}

template <class scalar>
bool SparseSortedMatrix<scalar>::SmallerRow(const int row1,const int row2)
{
  for (int j=0;j<cols();j++) {
    scalar result=coeff(row1,j)-coeff(row2,j);
    char sign=func::hardSign(result);
    if (sign!=0) return (sign<0);
  }
  return false;
}

template <class scalar>
bool SparseSortedMatrix<scalar>::LargerRow(const int row1,const int row2)
{
  for (int j=0;j<cols();j++) {
    scalar result=coeff(row1,j)-coeff(row2,j);
    char sign=func::hardSign(result);
    if (sign!=0) return (sign > 0);
  }
  return false;
}

template <class scalar>
void SparseSortedMatrix<scalar>::QuickAngleSort()
{
  std::multimap<refScalar,int> preorder;
  MatrixS centre=MatrixS::Zero(1,cols());
  for (int row=0;row<rows();row++) {
    sparseVector &data=(*this)[row];
    for (typename sparseVector::iterator it=data.begin();it!=data.end();it++) {
      centre.coeffRef(0,it->first)+=it->second;
    }
  }
  scalar count=rows();
  for (int col=0;col<cols();col++) centre.coeffRef(0,col)/=count;
  for (int row=0;row<rows();row++) {
    scalar angle=(coeff(row,1)-centre.coeff(0,1))/(coeff(row,0)-centre.coeff(0,0));
    angle=atan(angle);
    scalar dif=coeff(row,0)-centre.coeff(0,0);
    if (func::isNegative(dif)) angle+=func::const_pi(func::ms_1);
    preorder.insert(std::pair<refScalar,int>(func::toCentre(angle),row));
  }
  m_order.resize(rows());
  typename std::map<refScalar,int>::iterator it=preorder.begin();
  for (unsigned int row=0;row<rows() && row<preorder.size();row++,it++) {
    m_order[row]=it->second;
  }
}

template <class scalar>
void SparseSortedMatrix<scalar>::normalize()
{
  m_norms.resize(rows(),1);
  for (int row=0;row<rows();row++) {
    scalar value=coeff(row,0)*coeff(row,0);
    for (int col=1;col<cols();col++) func::madd(value,coeff(row,col),coeff(row,col));
    m_norms.coeffRef(row)=sqrt(value);
    sparseVector &data=(*this)[row];
    for (typename sparseVector::iterator it=data.begin();it!=data.end();it++) {
     it->second/=func::toCentre(m_norms.coeff(row));
    }
  }
}

template <class scalar>
void SparseSortedMatrix<scalar>::setRow(int row,int col,const MatrixS &src)
{
  sparseVector &data=(*this)[row];
  int size=cols()-col;
  for (int index=0;index<size;index++) {
    if (!func::isZero(src.coeff(0,index))) data[col+index]=src.coeff(0,index);
  }
}

template <class scalar>
void SparseSortedMatrix<scalar>::negRow(int row,int col,const MatrixS &src)
{
  sparseVector &data=(*this)[row];
  int size=cols()-col;
  for (int index=0;index<size;index++) {
    if (!func::isZero(src.coeff(0,index))) data[col+index]=-src.coeff(0,index);
  }
}

template <class scalar>
void SparseSortedMatrix<scalar>::subRow(MatrixS &dest,int row)
{
  sparseVector &data=(*this)[row];
  for (typename sparseVector::iterator it=data.begin();it!=data.end();it++) {
    dest.coeffRef(0,it->first)-=it->second;
  }
}

template <class scalar>
void SparseSortedMatrix<scalar>::cosineSort()
{
  int pos=1;
  int row=1;
  std::set<int> remaining;
  std::vector<refScalar> cosines;
  cosines.resize(rows()+1,func::ms_1);
  cosines[rows()]=0;
  m_order.resize(rows()+1);
  m_cosines.resize(rows()+1);
  m_cosines[pos]=func::ms_1;
  m_order[pos++]=row;
  for (int row=1;row<rows();row++) remaining.insert(row);

  while (pos<=rows()) {
    m_order[pos]=rows();  
    refScalar max=-func::ms_infinity;
    refScalar invCos=func::ms_1/cosines[row];
    refScalar lastCosine=func::ms_0;
    for (std::set<int>::iterator it=remaining.begin();it!=remaining.end();it++) {
      int row2=*it;
      if ((pos>2) && (!func::isNegative(lastCosine-cosines[row2]))) {
        if (cosines[row2]>func::ms_0) cosines[row2]*=invCos;
        else cosines[row2]+=func::ms_1-cosines[row];
        continue;
      }
      scalar value=coeff(row,0)*coeff(row2,0);
      for (int col=1;col<cols();col++) {
        func::madd(value,coeff(row,col),coeff(row2,col));
      }
      cosines[row2]=func::toCentre(value);
      if (func::toLower(value)>max) {
        max=func::toLower(value);
        m_cosines[pos]=max;
        m_order[pos]=row2;
        lastCosine=cosines[row2]*cosines[row];
      }
    }
    if (m_order[pos]==rows()) {
      if (remaining.size()<=0) break;
      int row2=*(remaining.begin());
      cosines[row2]=func::ms_1;
      continue;
    }
    row=m_order[pos++]++;
    remaining.erase(row);
  }
}

template <class scalar>
bool SparseSortedMatrix<scalar>::copyOrder(SparseSortedMatrix<scalar> &src)
{
  if (src.m_order.size()!=rows()) return false;
  m_order.clear();
  m_order.insert(m_order.begin(),src.m_order.begin(),src.m_order.end());
}

/// Pivots the tableau in place on the given pivot
template <class scalar>
void SparseSortedMatrix<scalar>::pivot(MatrixS &auxiliary,pivot_t pivot)
{
  int dimension=numCols();
  MatrixS epsMatrix(dimension,1);
  for (int col=0;col<numCols();col++) epsMatrix.coeffRef(col,0)=-getCoeff(pivot.row,col);
  epsMatrix.coeffRef(pivot.col,0)=func::ms_1;
  epsMatrix*=func::ms_1/((pivot.row>=numRows()) ? auxiliary.coeff(0,pivot.col) : getCoeff(pivot.row,pivot.col));
  m_dirty=false;
  for (int row = 0; row < numRows(); row++) setCoeff(row,pivot.col,entry(epsMatrix,row,0));
  if (auxiliary.cols()>0) auxiliary.coeffRef(0,pivot.col)=(auxiliary*epsMatrix).coeff(0,0);
  m_dirty=true;

/*  refScalar invXtemp0 = func::toCenter(func::ms_1/((pivot.row>=numRows()) ? auxiliary.coeff(0,pivot.col) : getCoeff(pivot.row,pivot.col)));
  scalar Xtemp;
  for (int row=0;row<numRows();row++) {
    if ((row!=pivot.row) && !func::isZero(getCoeff(row,pivot.col))) {
      Xtemp = coeff(row,pivot.col)*invXtemp0;
      for (int col = 0; col < numCols(); col++) {
        func::msub(coeffRef(row,col),coeff(row,col),Xtemp);
      }
    }
  }
  row(pivot.row)*=invXtemp0;
  if (auxiliary.cols()==numCols()) {
    Xtemp = auxiliary.coeff(0,pivot.col)*invXtemp0;
    for (int col = 0; col < numCols(); col++) {
      func::msub(auxiliary.coeffRef(0,col),auxiliary.coeff(0,col),Xtemp);
    }
  }*/
}

/// Retrieves the value of a tableau entry for a given row an column (see revised simplex algorithm)
template <class scalar>
scalar SparseSortedMatrix<scalar>::entry(MatrixS &basis,const int row, const int col)
{
  sparseVector &data=(*this)[row];
  if (m_dirty) return data[col];
  typename sparseVector::iterator it=data.begin();
  if (it==data.end()) return func::ms_0;
  scalar temp(it->second * basis.coeff(it->first,col));
  refScalar width,maxWidth=func::toWidth(temp);
  for (it++;it!=data.end();it++) {
    width=func::madd(temp,it->second,basis.coeff(it->first,col));
    if (width>maxWidth) maxWidth=width;
  }
  if (type::isInterval()) {
    char sign=func::hardSign(temp);
    if ((sign==0) && (maxWidth>0)) {
      maxWidth/=2;
      if ((func::toUpper(temp)<maxWidth) || (-func::toLower(temp)<maxWidth)) {
        maxWidth=maxTightWidth(basis,row,col);
        if ((func::toUpper(temp)<maxWidth) || (-func::toLower(temp)<maxWidth)) {
          if (maxWidth<func::ms_zero) {
            relaxBasis(basis,maxWidth,col);
            return tightEntry(basis,row,col);
          }
          func::imprecise(temp,maxWidth,"Imprecise width");
        }
      }
    }
  }
  return temp;
}

template <class scalar>
inline scalar SparseSortedMatrix<scalar>::tightEntry(MatrixS &basis,const int row, const int col,const int iteration)
/* Computes Tableau(row)*B^-1(col) */
{
  sparseVector &data=(*this)[row];
  typename sparseVector::iterator it=data.begin();
  if (it==data.end()) return func::ms_0;
  scalar temp(it->second * basis.coeff(it->first,col));
  refScalar width,maxWidth=func::tightWidth(it->second,basis.coeff(it->first,col));
  for (it++;it!=data.end();it++) {
    width=func::tightMadd(temp,it->second,basis.coeff(it->first,col));
    if (width>maxWidth) maxWidth=width;
  }
  if (type::isInterval()) {
    char sign=func::hardSign(temp);
    if (sign==0) {
      maxWidth/=2;
      maxWidth=maxTightWidth(basis,row,col);
      if ((func::toUpper(temp)<maxWidth) || (-func::toLower(temp)<maxWidth)) {
        if (maxWidth<func::ms_weakEpsilon) {
          relaxBasis(basis,maxWidth,col);
          return tightEntry(basis,row,col,iteration+1);
        }
        func::imprecise(temp,maxWidth,"tight width");
      }
    }
  }
  return temp;
}

template <class scalar>
inline scalar SparseSortedMatrix<scalar>::entry(const MatrixS &vector, const int row)
{ /*return the ith component of the vector  A x vector */
  sparseVector &data=(*this)[row];
  typename sparseVector::iterator it=data.begin();
  if (it==data.end()) return func::ms_0;
  scalar temp(it->second * vector.coeff(0,it->first));
  refScalar width,maxWidth=func::toWidth(temp);
  for (it++;it!=data.end();it++) {
    width=func::madd(temp,it->second,vector.coeff(0,it->first));
    if (width>maxWidth) maxWidth=width;
  }
  if (type::isInterval()) {
    char sign=func::hardSign(temp);
    if (sign==0) {
      maxWidth-=func::ms_epsilon;
      if ((func::toUpper(temp)<maxWidth) || (-func::toLower(temp)<maxWidth)) {
        maxWidth=maxTightWidth(vector,row);
        if ((func::toUpper(temp)<maxWidth) || (-func::toLower(temp)<maxWidth)) {
          /*if (maxWidth<func::ms_weakEpsilon) {
            scalar mult=func::ms_1+func::setpm(maxWidth);
            vector*=mult;
            return ATightValue(vector,row);
          }*/
          func::imprecise(temp,maxWidth,"vertex width");
        }
      }
    }
  }
  return temp;
}

template <class scalar>
inline scalar SparseSortedMatrix<scalar>::tightEntry(const MatrixS &vector, const long row)
{ /*return the ith component of the vector  A x vector */
  sparseVector &data=(*this)[row];
  typename sparseVector::iterator it=data.begin();
  if (it==data.end()) return func::ms_0;
  scalar temp(it->second * vector.coeff(0,it->first));
  refScalar width,maxWidth=func::tightWidth(it->second,vector.coeff(0,it->first));
  for (it++;it!=data.end();it++) {
    width=func::tightMadd(temp,it->second,vector.coeff(0,it->first));
    if (width>maxWidth) maxWidth=width;
  }
  if (type::isInterval()) {
    char sign=func::hardSign(temp);
    if (sign==0) {
      maxWidth/=2;
      if ((func::toUpper(temp)<maxWidth) || (-func::toLower(temp)<maxWidth)) {
        func::imprecise(temp,maxWidth,"tight vertex width");
      }
    }
  }
  return temp;
}

template <class scalar>
inline typename SparseSortedMatrix<scalar>::refScalar SparseSortedMatrix<scalar>::maxTightWidth(const MatrixS &vector, const int row)
{ /*return the minimum required width for full inclusion */
  sparseVector &data=(*this)[row];
  typename sparseVector::iterator it=data.begin();
  if (it==data.end()) return func::ms_0;
  refScalar width,maxWidth=func::tightWidth(it->second,vector.coeff(0,it->first));
  for (it++;it!=data.end();it++) {
    width=func::tightWidth(it->second,vector.coeff(0,it->first));
    if (width>maxWidth) maxWidth=width;
  }
  return maxWidth/2;
}

template <class scalar>
inline typename SparseSortedMatrix<scalar>::refScalar SparseSortedMatrix<scalar>::maxTightWidth(const MatrixS &basis,const int row, const int col)
/* Computes Tableau(row)*B^-1(col) */
{
  sparseVector &data=(*this)[row];
  typename sparseVector::iterator it=data.begin();
  if (it==data.end()) return func::ms_0;
  int nonzero=func::isZero(it->second*basis.coeff(it->first,col)) ? 0 : 1;
  refScalar width,maxWidth=func::tightWidth(it->second,basis.coeff(it->first,col));
  for (it++;it!=data.end();it++) {
    width=func::tightWidth(it->second,basis.coeff(it->first,col));
    nonzero+=func::isZero(it->second*basis.coeff(it->first,col)) ? 0 : 1;
    if (width>maxWidth) maxWidth=width;
  }
  if (nonzero<=1) return 0;
  return maxWidth/2;
}

///Relaxes the current basis by the given interval to cope with imprecisions
template <class scalar>
void SparseSortedMatrix<scalar>::relaxBasis(MatrixS &basis,scalar width,int col)
{
  if (!type::isInterval()) return;
  scalar mult=func::setpm(width);
  if (col>=0) {
    refScalar magOrder=10;
    for (int row=0;row<basis.rows();row++) {
      basis.coeffRef(row,col)+=mult*basis.coeff(row,col);
      basis.coeffRef(row,col)+=magOrder*(basis.coeff(row,col)-func::toCentre(basis.coeff(row,col)));
    }
  }
  else basis+=mult*basis;
}

#ifdef USE_LDOUBLE
  #ifdef USE_SINGLES
    template class SparseSortedMatrix<double_type>;
  #endif
  #ifdef USE_INTERVALS
    template class SparseSortedMatrix<ldinterval>;
  #endif
#endif
#ifdef USE_MPREAL
  #ifdef USE_SINGLES
    template class SparseSortedMatrix<mpfr::mpreal>;
  #endif
  #ifdef USE_INTERVALS
    template class SparseSortedMatrix<mpinterval>;
  #endif
#endif

}
